(* Operators *)

type op = Add | Sub | Mult | Div | Assign | Equal | Neq | Less | Leq | Greater | Geq
type quotes = Singleqt | Doubleqt
(*type paren = LPAREN | RPAREN*)


(* Literals *)

type lit = 
	Bool of bool
|	Int of int
|	Float of float
|	Char of quotes * char * quotes
|	String of quotes * string * quotes

let string_of_lit = function
	Bool(l) -> string_of_bool l
|	Int(l) -> string_of_int l
|	Float(l) -> string_of_float l
|	Char(l) ->
		(match l with Singleqt -> "\'" | Doubleqt -> "") ^
		l ^
		(match l with Singleqt -> "\'" | Doubleqt -> "")
|	String(l) ->
		(match l with Singleqt -> "" | Doubleqt -> "\"") ^
		l ^
		(match l with Singleqt -> "" | Doubleqt -> "\"")


(* Functions *)

type funcs = If | While | For | Print | Input
			| Fname of string

(* Expressions *)

type expr = 
	Literal of lit
|	Id of string
|	Call of (*string*)funcs * expr list
|	Oper of op * expr * expr
|	Noexpr
	

let rec string_of_expr = function
	Literal(l) -> string_of_lit l
|	Id(i) -> i
|	Call(f, elist) ->
		(match f with
			If -> "if" | While -> "while" | For -> "for" | Print -> "print" | Input -> "input" | Fname -> string
		) ^ 
		String.concat "" (List.map string_of_expr elist)
|	Oper(o, e1, e2) ->
		(match o with
			Add -> "+" | Sub -> "-" | Mult -> "*" | Div -> "/"
	  	 | 	Assign -> "=" | Equal -> "==" | Neq -> "!="
	  	 | 	Less -> "<" | Leq -> "<=" | Greater -> ">" | Geq -> ">="
		) ^
		string_of_expr e1 ^
		string_of_expr e2 ^
|	Noexpr -> ""
	
	
(* Programs *)

type program = 
	Prog of expr
	Set of expr list
	
let rec string_of_program = function
	Prog(e) -> string_of_expr e
	Set(s) -> String.concat "( " (List.map string_of_stmt s) ^ " )"
	
	
(*	
	
	
	
	(*Statement *)
type stmt =
	Stmt of expr
	Block of expr list
(*|	If of expr * stmt * stmt
|	While of expr * stmt
|	For of expr * sfor * expr list * stmt
|	Rules of expr * stmt
|	Assign of expr * stmt
|	Func of expr * stmt
|	Return of expr*)
|	Nostmt;;

let rec string_of_stmt = function
	Stmt(e) -> string_of_expr e
	
	Stmt(e) -> "( " ^ string_of_expr e ^ " )"
	Block(s) -> String.concat "( " (List.map string_of_stmt s) ^ " )"
|	Nostmt -> ""

let rec string_of_stmt = function
|	Stmt(e) -> String.concat "" (List.map string_of_expr exp) ^ "\n;"
|	Block(stmts) -> "{" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}"
|	If(exp1, stm1, stm2) -> string_of_expr exp1 ^ string_of_stmt stm1 ^ string_of_stmt stm2	
(* exp1 - Call, stmt1 and If (inside, stm1 - Block, stm2 - If or Nostmt)*)
|	While(exp, stm) -> string_of_expr exp ^ string_of_stmt stm
(* exp - Call, stmt - Block or Stmt *)
|	For(exp1,sf,exp2, stm) -> "for" ^ string_of_expr exp1 ^ "(" ^ String.concat (match sf with Colon -> "; " | NosepFor -> "") (List.map string_of_expr exp2) ^ ")" ^  string_of_stmt stm
(* exp1 - Use Sep as much as u want, with a NoExpr in between, exp2 - Op, stmt - Block or Stmt*)
|	Rules(exp, stm) -> "?" ^ string_of_expr exp ^ string_of_stmt stm
(* exp - Call or Noexpr, stmt - Block *)
|	Assign(exp, stm) -> string_of_expr exp ^ "=" ^ string_of_stmt stm
(* exp - Lit, stm - Stmt*)
|	Func(exp, stm) -> string_of_expr exp ^ string_of_stmt stm
(* exp - Lit, stm - Assign (inside, exp - Call or Noexp, stm - Block) *)
|	Return(exp) -> "return " ^ string_of_expr exp ^ ";\n";
(* exp - Op/Exp/Lit *)
|	Nostmt -> "";;
	
	
	










(* The Cardigan AST *)

(* Seperators *)
type sep = Space | Tab | Backslash | Nosep;;
type sfun = Comma | NosepFun;;
type sfor = Colon | NosepFor;;

(* Operators *)
type op = Add | Sub | Mult | Div | Power | Mod | NotEq | Eq | Greater | Less | Gte | Lte | And | Or | Inc | Dec | IncAdd | IncSub;;


(* Identifiers *)
type iden = string;;

(* Literals *)
type lit = 
	Bool of bool
|	Int of int
|	Float of float
|	String of string
|	Struct of iden * iden
|	Collection of iden * int
|	Enum of iden * iden;;

let string_of_lit = function
	Bool(l) -> string_of_bool l
|	Int(l) -> string_of_int l
|	Float(l) -> string_of_float l
|	String(l) -> l
|	Struct(i1, i2) -> i1 ^ "." ^ i2
|	Collection(i, n) -> i ^ "[" ^ string_of_int n ^ "]"
|	Enum(i1, i2) -> i1 ^ "." ^ i2;;


(* Expression *)
type expr =
	Sep of sep * expr * sep
|	Id of iden
|	Literal of lit
|	StructD of iden * expr
(*|	CollectionD of expr list*)
|	CollectionD of sfun * expr list
|	EnumD of sfun * expr list
|	Declare of string * expr
|	Op of expr * op * expr
|	Call of string * sfun * expr list
|	Noexpr;;

let rec string_of_expr = function
	Sep(s1, e, s2) ->
		(match s1 with Space -> " " | Tab -> "	" | Backslash -> "\\" | Nosep -> "") ^ 
		string_of_expr e ^
		(match s2 with Space -> " " | Tab -> "	" | Backslash -> "\\" | Nosep -> "")	
|	Id(s) -> s
|	Literal(l) -> string_of_lit l
|	StructD(i, e) -> i ^ ":" ^ string_of_expr e
(*|	CollectionD(e) ->
		"[" ^ String.concat ", " (List.map string_of_expr e) ^ "]"*)
|	CollectionD(sf,e) ->
		"[" ^ String.concat (match sf with Comma -> ", " | NosepFun -> "") (List.map string_of_expr e) ^ "]"
		
|	EnumD(sf, i) -> "|" ^ String.concat (match sf with Comma -> ", " | NosepFun -> "") (List.map string_of_expr i) ^ "|"
|	Declare(s, e) -> s ^ string_of_expr e 
|	Op(e1, op, e2) -> string_of_expr e1 ^
		(match op with Add -> "+" | Sub -> "-" | Mult -> "*" | Div -> "/" | Power -> "^" | Mod -> "%" | NotEq -> "!" | Eq -> "==" | Greater -> ">" | Less -> "<" | Gte -> ">=" | Lte -> "<=" | And -> " and " | Or -> " or " | Inc -> "++" | Dec -> "--" | IncAdd -> "+=" | IncSub -> "-=") ^
		string_of_expr e2
|	Call(f, sf, e) ->
(* CHECK IF THIS IS POSSIBLE sepFun = "" or ", " - FOR CALL*)
      f ^ "(" ^ String.concat (match sf with Comma -> ", " | NosepFun -> "") (List.map string_of_expr e) ^ ")"
|	Noexpr -> "";;


(*Statement *)
type stmt =
	Stmt of expr list
|	Block of stmt list
|	If of expr * stmt * stmt
|	While of expr * stmt
|	For of expr * sfor * expr list * stmt
|	Rules of expr * stmt
|	Assign of expr * stmt
|	Func of expr * stmt
|	Return of expr
|	Nostmt;;

let rec string_of_stmt = function
|	Stmt(exp) -> String.concat "" (List.map string_of_expr exp) ^ "\n;"
|	Block(stmts) -> "{" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}"
|	If(exp1, stm1, stm2) -> string_of_expr exp1 ^ string_of_stmt stm1 ^ string_of_stmt stm2	
(* exp1 - Call, stmt1 and If (inside, stm1 - Block, stm2 - If or Nostmt)*)
|	While(exp, stm) -> string_of_expr exp ^ string_of_stmt stm
(* exp - Call, stmt - Block or Stmt *)
|	For(exp1,sf,exp2, stm) -> "for" ^ string_of_expr exp1 ^ "(" ^ String.concat (match sf with Colon -> "; " | NosepFor -> "") (List.map string_of_expr exp2) ^ ")" ^  string_of_stmt stm
(* exp1 - Use Sep as much as u want, with a NoExpr in between, exp2 - Op, stmt - Block or Stmt*)
|	Rules(exp, stm) -> "?" ^ string_of_expr exp ^ string_of_stmt stm
(* exp - Call or Noexpr, stmt - Block *)
|	Assign(exp, stm) -> string_of_expr exp ^ "=" ^ string_of_stmt stm
(* exp - Lit, stm - Stmt*)
|	Func(exp, stm) -> string_of_expr exp ^ string_of_stmt stm
(* exp - Lit, stm - Assign (inside, exp - Call or Noexp, stm - Block) *)
|	Return(exp) -> "return " ^ string_of_expr exp ^ ";\n";
(* exp - Op/Exp/Lit *)
|	Nostmt -> "";;


(* The Final Program *)
type cardiganProgram = Program of stmt list;;

let runProgram = function
	Program(stmts) -> "PLAY" ^ String.concat "" (List.map string_of_stmt stmts);;
(* stmts start with assignments and then set of stmts *)


*)
